{-# OPTIONS -#include "SocketPrim_hsc.h" #-}
{-# LINE 1 "SocketPrim.hsc" #-}
-----------------------------------------------------------------------------
{-# LINE 2 "SocketPrim.hsc" #-}
-- $Id: SocketPrim.hsc,v 1.12.2.4 2001/12/13 17:14:24 simonmar Exp $
--
-- (c) The GRASP/AQUA Project, Glasgow University, 1995-1998,
--     The GHC Team 1999-2001
--
-- Low-level socket bindings

-- The SocketPrim module is for when you want full control over the
-- sockets, exposing the C socket API.

module SocketPrim (

    Socket,		-- instance Eq, Show
    Family(..),		
    SocketType(..),
    SockAddr(..),
    HostAddress,
    ShutdownCmd(..),
    ProtocolNumber,

    socket,		-- :: Family -> SocketType -> ProtocolNumber -> IO Socket 
    connect,		-- :: Socket -> SockAddr -> IO ()
    bindSocket,		-- :: Socket -> SockAddr -> IO ()
    listen,		-- :: Socket -> Int -> IO ()
    accept,		-- :: Socket -> IO (Socket, SockAddr)
    getPeerName,	-- :: Socket -> IO SockAddr
    getSocketName,	-- :: Socket -> IO SockAddr

    socketPort,		-- :: Socket -> IO PortNumber

    socketToHandle,	-- :: Socket -> IOMode -> IO Handle

    sendTo,		-- :: Socket -> String -> SockAddr -> IO Int
    recvFrom,		-- :: Socket -> Int -> IO (String, Int, SockAddr)
    
    send,		-- :: Socket -> String -> IO Int
    recv,		-- :: Socket -> Int    -> IO String

    inet_addr,		-- :: String -> IO HostAddress
    inet_ntoa,		-- :: HostAddress -> IO String

    sIsConnected,	-- :: Socket -> IO Bool
    sIsBound,		-- :: Socket -> IO Bool
    sIsListening,	-- :: Socket -> IO Bool 
    sIsReadable,	-- :: Socket -> IO Bool
    sIsWritable,	-- :: Socket -> IO Bool
    shutdown,		-- :: Socket -> ShutdownCmd -> IO ()
    sClose,		-- :: Socket -> IO ()

    -- socket opts
    SocketOption(..),
    getSocketOption,     -- :: Socket -> SocketOption -> IO Int
    setSocketOption,     -- :: Socket -> SocketOption -> Int -> IO ()

    PortNumber(..),	 -- instance (Eq, Ord, Enum, Num, Real, 
			 --		Integral, Storable)

    -- Special Constants
    aNY_PORT,		-- :: PortNumber
    iNADDR_ANY,		-- :: HostAddress
    sOMAXCONN,		-- :: Int
    maxListenQueue,	-- :: Int

    withSocketsDo,	-- :: IO a -> IO a

    -- The following are exported ONLY for use in the BSD module and
    -- should not be used anywhere else.

    packFamily, unpackFamily,
    packSocketType,

) where


{-# LINE 76 "SocketPrim.hsc" #-}

{-# LINE 77 "SocketPrim.hsc" #-}

import Foreign
import CForeign

import IO
import Monad ( liftM )
import Ratio ( (%) )


{-# LINE 86 "SocketPrim.hsc" #-}
import Bits
import MVar

import PrelConc		(threadWaitRead, threadWaitWrite)
import PrelHandle
import qualified PrelPosix
import PrelIOBase

{-# LINE 94 "SocketPrim.hsc" #-}

-----------------------------------------------------------------------------
-- Socket types

-- There are a few possible ways to do this.  The first is convert the
-- structs used in the C library into an equivalent Haskell type. An
-- other possible implementation is to keep all the internals in the C
-- code and use an Int## and a status flag. The second method is used
-- here since a lot of the C structures are not required to be
-- manipulated.

-- Originally the status was non-mutable so we had to return a new
-- socket each time we changed the status.  This version now uses
-- mutable variables to avoid the need to do this.  The result is a
-- cleaner interface and better security since the application
-- programmer now can't circumvent the status information to perform
-- invalid operations on sockets.

data SocketStatus
  -- Returned Status	Function called
  = NotConnected	-- socket
  | Bound		-- bindSocket
  | Listening		-- listen
  | Connected		-- connect/accept
    deriving (Eq, Show)

data Socket
  = MkSocket
	    CInt	         -- File Descriptor
	    Family				  
	    SocketType				  
	    ProtocolNumber	 -- Protocol Number
	    (MVar SocketStatus)  -- Status Flag

instance Eq Socket where
  (MkSocket _ _ _ _ m1) == (MkSocket _ _ _ _ m2) = m1 == m2

instance Show Socket where
  showsPrec n (MkSocket fd _ _ _ _) = 
	showString "<socket: " . shows fd . showString ">"

type ProtocolNumber = CInt

-- NOTE: HostAddresses are represented in network byte order.
--       Functions that expect the address in machine byte order
--       will have to perform the necessary translation.
type HostAddress = Word32

----------------------------------------------------------------------------
-- Port Numbers
--
-- newtyped to prevent accidental use of sane-looking
-- port numbers that haven't actually been converted to
-- network-byte-order first.
--
newtype PortNumber = PortNum Word16 deriving ( Eq, Ord )

instance Show PortNumber where
  showsPrec p pn = showsPrec p (portNumberToInt pn)

intToPortNumber :: Int -> PortNumber
intToPortNumber v = PortNum (htons (fromIntegral v))

portNumberToInt :: PortNumber -> Int
portNumberToInt (PortNum po) = fromIntegral (ntohs po)

foreign import "ntohs" unsafe ntohs :: Word16 -> Word16
foreign import "htons" unsafe htons :: Word16 -> Word16
foreign import "ntohl" unsafe ntohl :: Word32 -> Word32
foreign import "htonl" unsafe htonl :: Word32 -> Word32

instance Enum PortNumber where
    toEnum   = intToPortNumber
    fromEnum = portNumberToInt

instance Num PortNumber where
   fromInteger i = intToPortNumber (fromInteger i)
    -- for completeness.
   (+) x y   = intToPortNumber (portNumberToInt x + portNumberToInt y)
   (-) x y   = intToPortNumber (portNumberToInt x - portNumberToInt y)
   negate x  = intToPortNumber (-portNumberToInt x)
   (*) x y   = intToPortNumber (portNumberToInt x * portNumberToInt y)
   abs n     = intToPortNumber (abs (portNumberToInt n))
   signum n  = intToPortNumber (signum (portNumberToInt n))

instance Real PortNumber where
    toRational x = toInteger x % 1

instance Integral PortNumber where
    quotRem a b = let (c,d) = quotRem (portNumberToInt a) (portNumberToInt b) in
		  (intToPortNumber c, intToPortNumber d)
    toInteger a = toInteger (portNumberToInt a)

instance Storable PortNumber where
   sizeOf    _ = sizeOf    (undefined :: Word16)
   alignment _ = alignment (undefined :: Word16)
   poke p (PortNum po) = poke (castPtr p) po
   peek p = PortNum `liftM` peek (castPtr p)

-----------------------------------------------------------------------------
-- SockAddr

-- The scheme used for addressing sockets is somewhat quirky. The
-- calls in the BSD socket API that need to know the socket address
-- all operate in terms of struct sockaddr, a `virtual' type of
-- socket address.

-- The Internet family of sockets are addressed as struct sockaddr_in,
-- so when calling functions that operate on struct sockaddr, we have
-- to type cast the Internet socket address into a struct sockaddr.
-- Instances of the structure for different families might *not* be
-- the same size. Same casting is required of other families of
-- sockets such as Xerox NS. Similarly for Unix domain sockets.

-- To represent these socket addresses in Haskell-land, we do what BSD
-- didn't do, and use a union/algebraic type for the different
-- families. Currently only Unix domain sockets and the Internet family
-- are supported.

data SockAddr		-- C Names				
  = SockAddrInet
	PortNumber	-- sin_port  (network byte order)
	HostAddress	-- sin_addr  (ditto)

{-# LINE 218 "SocketPrim.hsc" #-}
  | SockAddrUnix
        String          -- sun_path

{-# LINE 221 "SocketPrim.hsc" #-}
  deriving (Eq)


{-# LINE 224 "SocketPrim.hsc" #-}
type CSaFamily = (Word16)
{-# LINE 225 "SocketPrim.hsc" #-}

{-# LINE 228 "SocketPrim.hsc" #-}

-- we can't write an instance of Storable for SockAddr, because the Storable
-- class can't easily handle alternatives.


{-# LINE 233 "SocketPrim.hsc" #-}
pokeSockAddr p (SockAddrUnix path) = do
	((\hsc_ptr -> pokeByteOff hsc_ptr 0)) p ((1) :: CSaFamily)
{-# LINE 235 "SocketPrim.hsc" #-}
	let pathC = map castCharToCChar path
	pokeArray0 0 (((\hsc_ptr -> hsc_ptr `plusPtr` 2)) p) pathC
{-# LINE 237 "SocketPrim.hsc" #-}

{-# LINE 238 "SocketPrim.hsc" #-}
pokeSockAddr p (SockAddrInet (PortNum port) addr) = do
	((\hsc_ptr -> pokeByteOff hsc_ptr 0)) p ((2) :: CSaFamily)
{-# LINE 240 "SocketPrim.hsc" #-}
	((\hsc_ptr -> pokeByteOff hsc_ptr 2)) p port
{-# LINE 241 "SocketPrim.hsc" #-}
	((\hsc_ptr -> pokeByteOff hsc_ptr 4)) p addr	
{-# LINE 242 "SocketPrim.hsc" #-}

peekSockAddr p = do
  family <- ((\hsc_ptr -> peekByteOff hsc_ptr 0)) p
{-# LINE 245 "SocketPrim.hsc" #-}
  case family :: CSaFamily of

{-# LINE 247 "SocketPrim.hsc" #-}
	(1) -> do
{-# LINE 248 "SocketPrim.hsc" #-}
		str <- peekCString (((\hsc_ptr -> hsc_ptr `plusPtr` 2)) p)
{-# LINE 249 "SocketPrim.hsc" #-}
		return (SockAddrUnix str)

{-# LINE 251 "SocketPrim.hsc" #-}
	(2) -> do
{-# LINE 252 "SocketPrim.hsc" #-}
		addr <- ((\hsc_ptr -> peekByteOff hsc_ptr 4)) p
{-# LINE 253 "SocketPrim.hsc" #-}
		port <- ((\hsc_ptr -> peekByteOff hsc_ptr 2)) p
{-# LINE 254 "SocketPrim.hsc" #-}
		return (SockAddrInet (PortNum port) addr)

-- size of struct sockaddr by family

{-# LINE 258 "SocketPrim.hsc" #-}
sizeOfSockAddr_Family AF_UNIX = 110
{-# LINE 259 "SocketPrim.hsc" #-}

{-# LINE 260 "SocketPrim.hsc" #-}
sizeOfSockAddr_Family AF_INET = 16
{-# LINE 261 "SocketPrim.hsc" #-}

-- size of struct sockaddr by SockAddr

{-# LINE 264 "SocketPrim.hsc" #-}
sizeOfSockAddr (SockAddrUnix _)   = 110
{-# LINE 265 "SocketPrim.hsc" #-}

{-# LINE 266 "SocketPrim.hsc" #-}
sizeOfSockAddr (SockAddrInet _ _) = 16
{-# LINE 267 "SocketPrim.hsc" #-}

withSockAddr :: SockAddr -> (Ptr SockAddr -> Int -> IO a) -> IO a
withSockAddr addr f = do
 let sz = sizeOfSockAddr addr
 allocaBytes sz $ \p -> pokeSockAddr p addr >> f (castPtr p) sz

withNewSockAddr :: Family -> (Ptr SockAddr -> Int -> IO a) -> IO a
withNewSockAddr family f = do
 let sz = sizeOfSockAddr_Family family
 allocaBytes sz $ \ptr -> f ptr sz

-----------------------------------------------------------------------------
-- Connection Functions

-- In the following connection and binding primitives.  The names of
-- the equivalent C functions have been preserved where possible. It
-- should be noted that some of these names used in the C library,
-- \tr{bind} in particular, have a different meaning to many Haskell
-- programmers and have thus been renamed by appending the prefix
-- Socket.

-- Create an unconnected socket of the given family, type and
-- protocol.  The most common invocation of $socket$ is the following:
--    ...
--    my_socket <- socket AF_INET Stream 6
--    ...

socket :: Family 	 -- Family Name (usually AF_INET)
       -> SocketType 	 -- Socket Type (usually Stream)
       -> ProtocolNumber -- Protocol Number (getProtocolByName to find value)
       -> IO Socket	 -- Unconnected Socket

socket family stype protocol = do
    fd <- throwErrnoIfMinus1Retry "socket" $
		c_socket (packFamily family) (packSocketType stype) protocol
    setSocketNonBlocking fd
    socket_status <- newMVar NotConnected
    return (MkSocket fd family stype protocol socket_status)
      
-----------------------------------------------------------------------------
-- Binding a socket
--
-- Given a port number this {\em binds} the socket to that port. This
-- means that the programmer is only interested in data being sent to
-- that port number. The $Family$ passed to $bindSocket$ must
-- be the same as that passed to $socket$.	 If the special port
-- number $aNY\_PORT$ is passed then the system assigns the next
-- available use port.
-- 
-- Port numbers for standard unix services can be found by calling
-- $getServiceEntry$.  These are traditionally port numbers below
-- 1000; although there are afew, namely NFS and IRC, which used higher
-- numbered ports.
-- 
-- The port number allocated to a socket bound by using $aNY\_PORT$ can be
-- found by calling $port$

bindSocket :: Socket	-- Unconnected Socket
	   -> SockAddr	-- Address to Bind to
	   -> IO ()

bindSocket (MkSocket s _family _stype _protocol socketStatus) addr = do
 modifyMVar_ socketStatus $ \ status -> do
 if status /= NotConnected 
  then
   ioError (userError ("bindSocket: can't peform bind on socket in status " ++
	 show status))
  else do
   withSockAddr addr $ \p_addr sz -> do
   status <- throwErrnoIfMinus1Retry "bind" $ c_bind s p_addr (fromIntegral sz)
   return Bound

-----------------------------------------------------------------------------
-- Connecting a socket
--
-- Make a connection to an already opened socket on a given machine
-- and port.  assumes that we have already called createSocket,
-- otherwise it will fail.
--
-- This is the dual to $bindSocket$.  The {\em server} process will
-- usually bind to a port number, the {\em client} will then connect
-- to the same port number.  Port numbers of user applications are
-- normally agreed in advance, otherwise we must rely on some meta
-- protocol for telling the other side what port number we have been
-- allocated.

connect :: Socket	-- Unconnected Socket
	-> SockAddr 	-- Socket address stuff
	-> IO ()

connect sock@(MkSocket s _family _stype _protocol socketStatus) addr = do
 modifyMVar_ socketStatus $ \currentStatus -> do
 if currentStatus /= NotConnected 
  then
   ioError (userError ("connect: can't peform connect on socket in status " ++
         show currentStatus))
  else do
   withSockAddr addr $ \p_addr sz -> do

   let  connectLoop = do
       	   r <- c_connect s p_addr (fromIntegral sz)
       	   if r == -1
       	       then do err <- getErrno
		       case () of
			 _ | err == eINTR       -> connectLoop
			 _ | err == eINPROGRESS -> connectBlocked
			 otherwise              -> throwErrno "connect"
       	       else return r

	connectBlocked = do 
	   threadWaitWrite (fromIntegral s)
	   err <- getSocketOption sock SoError
	   if (err == 0)
	   	then return 0
	   	else do ioError (errnoToIOError "connect" 
	   			(Errno (fromIntegral err))
	   			Nothing Nothing)

   connectLoop
   return Connected

-----------------------------------------------------------------------------
-- Listen
--
-- The programmer must call $listen$ to tell the system software that
-- they are now interested in receiving data on this port.  This must
-- be called on the bound socket before any calls to read or write
-- data are made.

-- The programmer also gives a number which indicates the length of
-- the incoming queue of unread messages for this socket. On most
-- systems the maximum queue length is around 5.  To remove a message
-- from the queue for processing a call to $accept$ should be made.

listen :: Socket  -- Connected & Bound Socket
       -> Int 	  -- Queue Length
       -> IO ()

listen (MkSocket s _family _stype _protocol socketStatus) backlog = do
 modifyMVar_ socketStatus $ \ status -> do
 if status /= Bound 
   then
    ioError (userError ("listen: can't peform listen on socket in status " ++
          show status))
   else do
    throwErrnoIfMinus1Retry "listen" (c_listen s (fromIntegral backlog))
    return Listening

-----------------------------------------------------------------------------
-- Accept
--
-- A call to `accept' only returns when data is available on the given
-- socket, unless the socket has been set to non-blocking.  It will
-- return a new socket which should be used to read the incoming data and
-- should then be closed. Using the socket returned by `accept' allows
-- incoming requests to be queued on the original socket.

accept :: Socket			-- Queue Socket
       -> IO (Socket,			-- Readable Socket
	      SockAddr)			-- Peer details

accept sock@(MkSocket s family stype protocol status) = do
 currentStatus <- readMVar status
 okay <- sIsAcceptable sock
 if not okay
   then
     ioError (userError ("accept: can't peform accept on socket in status " ++
	 show currentStatus))
   else do
     let sz = sizeOfSockAddr_Family family
     allocaBytes sz $ \ sockaddr -> do
     withObject (fromIntegral sz) $ \ ptr_len -> do
     new_sock <- throwErrnoIfMinus1Retry_repeatOnBlock "accept" 
			(threadWaitRead (fromIntegral s))
			(c_accept s sockaddr ptr_len)
     addr <- peekSockAddr sockaddr
     new_status <- newMVar Connected
     return ((MkSocket new_sock family stype protocol new_status), addr)

-----------------------------------------------------------------------------
-- sendTo & recvFrom

sendTo :: Socket	-- (possibly) bound/connected Socket
       -> String	-- Data to send
       -> SockAddr
       -> IO Int	-- Number of Bytes sent

sendTo (MkSocket s _family _stype _protocol status) xs addr = do
 withSockAddr addr $ \p_addr sz -> do
 withCString xs $ \str -> do
   liftM fromIntegral $
     throwErrnoIfMinus1Retry_repeatOnBlock "sendTo"
	(threadWaitWrite (fromIntegral s)) $
	c_sendto s str (fromIntegral $ length xs) 0{-flags-} 
			p_addr (fromIntegral sz)


recvFrom :: Socket -> Int -> IO (String, Int, SockAddr)
recvFrom sock@(MkSocket s _family _stype _protocol status) nbytes = do
  allocaBytes nbytes $ \ptr -> do
    withNewSockAddr AF_INET $ \ptr_addr sz -> do
      alloca $ \ptr_len -> do
      	poke ptr_len (fromIntegral sz)
        len <- throwErrnoIfMinus1Retry_repeatOnBlock "recvFrom" 
        	   (threadWaitRead (fromIntegral s)) $
        	   c_recvfrom s ptr (fromIntegral nbytes) 0{-flags-} 
				ptr_addr ptr_len
        let len' = fromIntegral len
	flg <- sIsConnected sock
	  -- For at least one implementation (WinSock 2), recvfrom() ignores
	  -- filling in the sockaddr for connected TCP sockets. Cope with 
	  -- this by using getPeerName instead.
	sockaddr <- 
		if flg then
		   getPeerName sock
		else
		   peekSockAddr ptr_addr 
        str <- peekCStringLen (ptr,len')
        return (str, len', sockaddr)

-----------------------------------------------------------------------------
-- send & recv

send :: Socket	-- Bound/Connected Socket
     -> String	-- Data to send
     -> IO Int	-- Number of Bytes sent
send (MkSocket s _family _stype _protocol status) xs = do
 withCString xs $ \str -> do
   liftM fromIntegral $
     throwErrnoIfMinus1Retry_repeatOnBlock "send"
	(threadWaitWrite (fromIntegral s)) $
	c_send s str (fromIntegral $ length xs) 0{-flags-} 

recv :: Socket -> Int -> IO String
recv sock@(MkSocket s _family _stype _protocol status) nbytes = do
  allocaBytes nbytes $ \ptr -> do
        len <- throwErrnoIfMinus1Retry_repeatOnBlock "recv" 
        	   (threadWaitRead (fromIntegral s)) $
        	   c_recv s ptr (fromIntegral nbytes) 0{-flags-} 
        let len' = fromIntegral len
        peekCStringLen (ptr,len')

-- ---------------------------------------------------------------------------
-- socketPort
--
-- The port number the given socket is currently connected to can be
-- determined by calling $port$, is generally only useful when bind
-- was given $aNY\_PORT$.

socketPort :: Socket		-- Connected & Bound Socket
	   -> IO PortNumber	-- Port Number of Socket
socketPort sock@(MkSocket _ AF_INET _ _ _) = do
    (SockAddrInet port _) <- getSocketName sock
    return port
socketPort (MkSocket _ family _ _ _) =
    ioError (userError ("socketPort: not supported for Family " ++ show family))


-- ---------------------------------------------------------------------------
-- getPeerName

-- Calling $getPeerName$ returns the address details of the machine,
-- other than the local one, which is connected to the socket. This is
-- used in programs such as FTP to determine where to send the
-- returning data.  The corresponding call to get the details of the
-- local machine is $getSocketName$.

getPeerName   :: Socket -> IO SockAddr
getPeerName (MkSocket s family _ _ _) = do
 withNewSockAddr family $ \ptr sz -> do
   withObject (fromIntegral sz) $ \int_star -> do
   throwErrnoIfMinus1Retry "getPeerName" $ c_getpeername s ptr int_star
   sz <- peek int_star
   peekSockAddr ptr
    
getSocketName :: Socket -> IO SockAddr
getSocketName (MkSocket s family _ _ _) = do
 withNewSockAddr family $ \ptr sz -> do
   withObject (fromIntegral sz) $ \int_star -> do
   throwErrnoIfMinus1Retry "getSocketName" $ c_getsockname s ptr int_star
   peekSockAddr ptr

-----------------------------------------------------------------------------
-- Socket Properties

data SocketOption
    = DummySocketOption__

{-# LINE 555 "SocketPrim.hsc" #-}
    | Debug         {- SO_DEBUG     -}

{-# LINE 557 "SocketPrim.hsc" #-}

{-# LINE 558 "SocketPrim.hsc" #-}
    | ReuseAddr     {- SO_REUSEADDR -}

{-# LINE 560 "SocketPrim.hsc" #-}

{-# LINE 561 "SocketPrim.hsc" #-}
    | Type          {- SO_TYPE      -}

{-# LINE 563 "SocketPrim.hsc" #-}

{-# LINE 564 "SocketPrim.hsc" #-}
    | SoError       {- SO_ERROR     -}

{-# LINE 566 "SocketPrim.hsc" #-}

{-# LINE 567 "SocketPrim.hsc" #-}
    | DontRoute     {- SO_DONTROUTE -}

{-# LINE 569 "SocketPrim.hsc" #-}

{-# LINE 570 "SocketPrim.hsc" #-}
    | Broadcast     {- SO_BROADCAST -}

{-# LINE 572 "SocketPrim.hsc" #-}

{-# LINE 573 "SocketPrim.hsc" #-}
    | SendBuffer    {- SO_SNDBUF    -}

{-# LINE 575 "SocketPrim.hsc" #-}

{-# LINE 576 "SocketPrim.hsc" #-}
    | RecvBuffer    {- SO_RCVBUF    -}

{-# LINE 578 "SocketPrim.hsc" #-}

{-# LINE 579 "SocketPrim.hsc" #-}
    | KeepAlive     {- SO_KEEPALIVE -}

{-# LINE 581 "SocketPrim.hsc" #-}

{-# LINE 582 "SocketPrim.hsc" #-}
    | OOBInline     {- SO_OOBINLINE -}

{-# LINE 584 "SocketPrim.hsc" #-}

{-# LINE 585 "SocketPrim.hsc" #-}
    | MaxSegment    {- TCP_MAXSEG   -}

{-# LINE 587 "SocketPrim.hsc" #-}

{-# LINE 588 "SocketPrim.hsc" #-}
    | NoDelay       {- TCP_NODELAY  -}

{-# LINE 590 "SocketPrim.hsc" #-}

{-# LINE 591 "SocketPrim.hsc" #-}
    | Linger        {- SO_LINGER    -}

{-# LINE 593 "SocketPrim.hsc" #-}

{-# LINE 596 "SocketPrim.hsc" #-}

{-# LINE 597 "SocketPrim.hsc" #-}
    | RecvLowWater  {- SO_RCVLOWAT  -}

{-# LINE 599 "SocketPrim.hsc" #-}

{-# LINE 600 "SocketPrim.hsc" #-}
    | SendLowWater  {- SO_SNDLOWAT  -}

{-# LINE 602 "SocketPrim.hsc" #-}

{-# LINE 603 "SocketPrim.hsc" #-}
    | RecvTimeOut   {- SO_RCVTIMEO  -}

{-# LINE 605 "SocketPrim.hsc" #-}

{-# LINE 606 "SocketPrim.hsc" #-}
    | SendTimeOut   {- SO_SNDTIMEO  -}

{-# LINE 608 "SocketPrim.hsc" #-}

{-# LINE 611 "SocketPrim.hsc" #-}

socketOptLevel :: SocketOption -> CInt
socketOptLevel so = 
  case so of

{-# LINE 616 "SocketPrim.hsc" #-}
    MaxSegment   -> 6
{-# LINE 617 "SocketPrim.hsc" #-}

{-# LINE 618 "SocketPrim.hsc" #-}

{-# LINE 619 "SocketPrim.hsc" #-}
    NoDelay      -> 6
{-# LINE 620 "SocketPrim.hsc" #-}

{-# LINE 621 "SocketPrim.hsc" #-}
    _            -> 1
{-# LINE 622 "SocketPrim.hsc" #-}

packSocketOption :: SocketOption -> CInt
packSocketOption so =
  case so of

{-# LINE 627 "SocketPrim.hsc" #-}
    Debug         -> 1
{-# LINE 628 "SocketPrim.hsc" #-}

{-# LINE 629 "SocketPrim.hsc" #-}

{-# LINE 630 "SocketPrim.hsc" #-}
    ReuseAddr     -> 2
{-# LINE 631 "SocketPrim.hsc" #-}

{-# LINE 632 "SocketPrim.hsc" #-}

{-# LINE 633 "SocketPrim.hsc" #-}
    Type          -> 3
{-# LINE 634 "SocketPrim.hsc" #-}

{-# LINE 635 "SocketPrim.hsc" #-}

{-# LINE 636 "SocketPrim.hsc" #-}
    SoError       -> 4
{-# LINE 637 "SocketPrim.hsc" #-}

{-# LINE 638 "SocketPrim.hsc" #-}

{-# LINE 639 "SocketPrim.hsc" #-}
    DontRoute     -> 5
{-# LINE 640 "SocketPrim.hsc" #-}

{-# LINE 641 "SocketPrim.hsc" #-}

{-# LINE 642 "SocketPrim.hsc" #-}
    Broadcast     -> 6
{-# LINE 643 "SocketPrim.hsc" #-}

{-# LINE 644 "SocketPrim.hsc" #-}

{-# LINE 645 "SocketPrim.hsc" #-}
    SendBuffer    -> 7
{-# LINE 646 "SocketPrim.hsc" #-}

{-# LINE 647 "SocketPrim.hsc" #-}

{-# LINE 648 "SocketPrim.hsc" #-}
    RecvBuffer    -> 8
{-# LINE 649 "SocketPrim.hsc" #-}

{-# LINE 650 "SocketPrim.hsc" #-}

{-# LINE 651 "SocketPrim.hsc" #-}
    KeepAlive     -> 9
{-# LINE 652 "SocketPrim.hsc" #-}

{-# LINE 653 "SocketPrim.hsc" #-}

{-# LINE 654 "SocketPrim.hsc" #-}
    OOBInline     -> 10
{-# LINE 655 "SocketPrim.hsc" #-}

{-# LINE 656 "SocketPrim.hsc" #-}

{-# LINE 657 "SocketPrim.hsc" #-}
    MaxSegment    -> 2
{-# LINE 658 "SocketPrim.hsc" #-}

{-# LINE 659 "SocketPrim.hsc" #-}

{-# LINE 660 "SocketPrim.hsc" #-}
    NoDelay       -> 1
{-# LINE 661 "SocketPrim.hsc" #-}

{-# LINE 662 "SocketPrim.hsc" #-}

{-# LINE 665 "SocketPrim.hsc" #-}

{-# LINE 666 "SocketPrim.hsc" #-}
    RecvLowWater  -> 18
{-# LINE 667 "SocketPrim.hsc" #-}

{-# LINE 668 "SocketPrim.hsc" #-}

{-# LINE 669 "SocketPrim.hsc" #-}
    SendLowWater  -> 19
{-# LINE 670 "SocketPrim.hsc" #-}

{-# LINE 671 "SocketPrim.hsc" #-}

{-# LINE 672 "SocketPrim.hsc" #-}
    RecvTimeOut   -> 20
{-# LINE 673 "SocketPrim.hsc" #-}

{-# LINE 674 "SocketPrim.hsc" #-}

{-# LINE 675 "SocketPrim.hsc" #-}
    SendTimeOut   -> 21
{-# LINE 676 "SocketPrim.hsc" #-}

{-# LINE 677 "SocketPrim.hsc" #-}

{-# LINE 680 "SocketPrim.hsc" #-}

setSocketOption :: Socket 
		-> SocketOption -- Option Name
		-> Int		-- Option Value
		-> IO ()
setSocketOption (MkSocket s _ _ _ _) so v = do
   withObject (fromIntegral v) $ \ptr_v -> do
   throwErrnoIfMinus1_ "setSocketOption" $
       c_setsockopt s (socketOptLevel so) (packSocketOption so) ptr_v 
	  (fromIntegral (sizeOf v))
   return ()


getSocketOption :: Socket
		-> SocketOption  -- Option Name
		-> IO Int	 -- Option Value
getSocketOption (MkSocket s _ _ _ _) so = do
   alloca $ \ptr_v ->
     withObject (fromIntegral (sizeOf (undefined :: CInt))) $ \ptr_sz -> do
       throwErrnoIfMinus1 "getSocketOption" $
	 c_getsockopt s (socketOptLevel so) (packSocketOption so) ptr_v ptr_sz
       fromIntegral `liftM` peek ptr_v

{-
A calling sequence table for the main functions is shown in the table below.

\begin{figure}[h]
\begin{center}
\begin{tabular}{|l|c|c|c|c|c|c|c|}d
\hline
{\bf A Call to} & socket & connect & bindSocket & listen & accept & read & write \\
\hline
{\bf Precedes} & & & & & & & \\
\hline 
socket &	&	  &	       &	&	 &	& \\
\hline
connect & +	&	  &	       &	&	 &	& \\
\hline
bindSocket & +	&	  &	       &	&	 &	& \\
\hline
listen &	&	  & +	       &	&	 &	& \\
\hline
accept &	&	  &	       &  +	&	 &	& \\
\hline
read   &	&   +	  &	       &  +	&  +	 &  +	& + \\
\hline
write  &	&   +	  &	       &  +	&  +	 &  +	& + \\
\hline
\end{tabular}
\caption{Sequence Table for Major functions of Socket}
\label{tab:api-seq}
\end{center}
\end{figure}
-}

-- ---------------------------------------------------------------------------
-- OS Dependent Definitions
    
unpackFamily	:: CInt -> Family
packFamily	:: Family -> CInt

packSocketType	:: SocketType -> CInt

data Family
	= AF_UNSPEC	-- unspecified

{-# LINE 746 "SocketPrim.hsc" #-}
	| AF_UNIX	-- local to host (pipes, portals

{-# LINE 748 "SocketPrim.hsc" #-}

{-# LINE 749 "SocketPrim.hsc" #-}
	| AF_INET	-- internetwork: UDP, TCP, etc

{-# LINE 751 "SocketPrim.hsc" #-}

{-# LINE 752 "SocketPrim.hsc" #-}
        | AF_INET6	-- Internet Protocol version 6

{-# LINE 754 "SocketPrim.hsc" #-}

{-# LINE 757 "SocketPrim.hsc" #-}

{-# LINE 760 "SocketPrim.hsc" #-}

{-# LINE 763 "SocketPrim.hsc" #-}

{-# LINE 766 "SocketPrim.hsc" #-}

{-# LINE 769 "SocketPrim.hsc" #-}

{-# LINE 772 "SocketPrim.hsc" #-}

{-# LINE 775 "SocketPrim.hsc" #-}

{-# LINE 778 "SocketPrim.hsc" #-}

{-# LINE 779 "SocketPrim.hsc" #-}
	| AF_SNA	-- IBM SNA

{-# LINE 781 "SocketPrim.hsc" #-}

{-# LINE 782 "SocketPrim.hsc" #-}
	| AF_DECnet	-- DECnet

{-# LINE 784 "SocketPrim.hsc" #-}

{-# LINE 787 "SocketPrim.hsc" #-}

{-# LINE 790 "SocketPrim.hsc" #-}

{-# LINE 793 "SocketPrim.hsc" #-}

{-# LINE 794 "SocketPrim.hsc" #-}
	| AF_APPLETALK	-- Apple Talk

{-# LINE 796 "SocketPrim.hsc" #-}

{-# LINE 797 "SocketPrim.hsc" #-}
	| AF_ROUTE	-- Internal Routing Protocol 

{-# LINE 799 "SocketPrim.hsc" #-}

{-# LINE 802 "SocketPrim.hsc" #-}

{-# LINE 805 "SocketPrim.hsc" #-}

{-# LINE 808 "SocketPrim.hsc" #-}

{-# LINE 811 "SocketPrim.hsc" #-}

{-# LINE 814 "SocketPrim.hsc" #-}

{-# LINE 817 "SocketPrim.hsc" #-}

{-# LINE 818 "SocketPrim.hsc" #-}
	| AF_X25	-- CCITT X.25

{-# LINE 820 "SocketPrim.hsc" #-}

{-# LINE 821 "SocketPrim.hsc" #-}
	| AF_AX25

{-# LINE 823 "SocketPrim.hsc" #-}

{-# LINE 826 "SocketPrim.hsc" #-}

{-# LINE 829 "SocketPrim.hsc" #-}

{-# LINE 830 "SocketPrim.hsc" #-}
	| AF_IPX	-- Novell Internet Protocol

{-# LINE 832 "SocketPrim.hsc" #-}

{-# LINE 835 "SocketPrim.hsc" #-}

{-# LINE 838 "SocketPrim.hsc" #-}

{-# LINE 841 "SocketPrim.hsc" #-}

{-# LINE 844 "SocketPrim.hsc" #-}

{-# LINE 847 "SocketPrim.hsc" #-}

{-# LINE 850 "SocketPrim.hsc" #-}

{-# LINE 853 "SocketPrim.hsc" #-}

{-# LINE 856 "SocketPrim.hsc" #-}

{-# LINE 859 "SocketPrim.hsc" #-}

{-# LINE 862 "SocketPrim.hsc" #-}

{-# LINE 865 "SocketPrim.hsc" #-}

{-# LINE 868 "SocketPrim.hsc" #-}

{-# LINE 871 "SocketPrim.hsc" #-}

{-# LINE 874 "SocketPrim.hsc" #-}

{-# LINE 877 "SocketPrim.hsc" #-}

{-# LINE 880 "SocketPrim.hsc" #-}

{-# LINE 883 "SocketPrim.hsc" #-}

{-# LINE 886 "SocketPrim.hsc" #-}

{-# LINE 889 "SocketPrim.hsc" #-}

{-# LINE 892 "SocketPrim.hsc" #-}

{-# LINE 895 "SocketPrim.hsc" #-}
	deriving (Eq, Ord, Show)

------ ------
			
packFamily f = case f of
	AF_UNSPEC -> 0
{-# LINE 901 "SocketPrim.hsc" #-}

{-# LINE 902 "SocketPrim.hsc" #-}
	AF_UNIX -> 1
{-# LINE 903 "SocketPrim.hsc" #-}

{-# LINE 904 "SocketPrim.hsc" #-}

{-# LINE 905 "SocketPrim.hsc" #-}
	AF_INET -> 2
{-# LINE 906 "SocketPrim.hsc" #-}

{-# LINE 907 "SocketPrim.hsc" #-}

{-# LINE 908 "SocketPrim.hsc" #-}
        AF_INET6 -> 10
{-# LINE 909 "SocketPrim.hsc" #-}

{-# LINE 910 "SocketPrim.hsc" #-}

{-# LINE 913 "SocketPrim.hsc" #-}

{-# LINE 916 "SocketPrim.hsc" #-}

{-# LINE 919 "SocketPrim.hsc" #-}

{-# LINE 922 "SocketPrim.hsc" #-}

{-# LINE 925 "SocketPrim.hsc" #-}

{-# LINE 928 "SocketPrim.hsc" #-}

{-# LINE 931 "SocketPrim.hsc" #-}

{-# LINE 934 "SocketPrim.hsc" #-}

{-# LINE 935 "SocketPrim.hsc" #-}
	AF_SNA -> 22
{-# LINE 936 "SocketPrim.hsc" #-}

{-# LINE 937 "SocketPrim.hsc" #-}

{-# LINE 938 "SocketPrim.hsc" #-}
	AF_DECnet -> 12
{-# LINE 939 "SocketPrim.hsc" #-}

{-# LINE 940 "SocketPrim.hsc" #-}

{-# LINE 943 "SocketPrim.hsc" #-}

{-# LINE 946 "SocketPrim.hsc" #-}

{-# LINE 949 "SocketPrim.hsc" #-}

{-# LINE 950 "SocketPrim.hsc" #-}
	AF_APPLETALK -> 5
{-# LINE 951 "SocketPrim.hsc" #-}

{-# LINE 952 "SocketPrim.hsc" #-}

{-# LINE 953 "SocketPrim.hsc" #-}
	AF_ROUTE -> 16
{-# LINE 954 "SocketPrim.hsc" #-}

{-# LINE 955 "SocketPrim.hsc" #-}

{-# LINE 958 "SocketPrim.hsc" #-}

{-# LINE 961 "SocketPrim.hsc" #-}

{-# LINE 964 "SocketPrim.hsc" #-}

{-# LINE 967 "SocketPrim.hsc" #-}

{-# LINE 970 "SocketPrim.hsc" #-}

{-# LINE 973 "SocketPrim.hsc" #-}

{-# LINE 974 "SocketPrim.hsc" #-}
	AF_X25 -> 9
{-# LINE 975 "SocketPrim.hsc" #-}

{-# LINE 976 "SocketPrim.hsc" #-}

{-# LINE 977 "SocketPrim.hsc" #-}
	AF_AX25 -> 3
{-# LINE 978 "SocketPrim.hsc" #-}

{-# LINE 979 "SocketPrim.hsc" #-}

{-# LINE 982 "SocketPrim.hsc" #-}

{-# LINE 985 "SocketPrim.hsc" #-}

{-# LINE 986 "SocketPrim.hsc" #-}
	AF_IPX -> 4
{-# LINE 987 "SocketPrim.hsc" #-}

{-# LINE 988 "SocketPrim.hsc" #-}

{-# LINE 991 "SocketPrim.hsc" #-}

{-# LINE 994 "SocketPrim.hsc" #-}

{-# LINE 997 "SocketPrim.hsc" #-}

{-# LINE 1000 "SocketPrim.hsc" #-}

{-# LINE 1003 "SocketPrim.hsc" #-}

{-# LINE 1006 "SocketPrim.hsc" #-}

{-# LINE 1009 "SocketPrim.hsc" #-}

{-# LINE 1012 "SocketPrim.hsc" #-}

{-# LINE 1015 "SocketPrim.hsc" #-}

{-# LINE 1018 "SocketPrim.hsc" #-}

{-# LINE 1021 "SocketPrim.hsc" #-}

{-# LINE 1024 "SocketPrim.hsc" #-}

{-# LINE 1027 "SocketPrim.hsc" #-}

{-# LINE 1030 "SocketPrim.hsc" #-}

{-# LINE 1033 "SocketPrim.hsc" #-}

{-# LINE 1036 "SocketPrim.hsc" #-}

{-# LINE 1039 "SocketPrim.hsc" #-}

{-# LINE 1042 "SocketPrim.hsc" #-}

{-# LINE 1045 "SocketPrim.hsc" #-}

{-# LINE 1048 "SocketPrim.hsc" #-}

{-# LINE 1051 "SocketPrim.hsc" #-}

--------- ----------

unpackFamily f = case f of
	(0) -> AF_UNSPEC
{-# LINE 1056 "SocketPrim.hsc" #-}

{-# LINE 1057 "SocketPrim.hsc" #-}
	(1) -> AF_UNIX
{-# LINE 1058 "SocketPrim.hsc" #-}

{-# LINE 1059 "SocketPrim.hsc" #-}

{-# LINE 1060 "SocketPrim.hsc" #-}
	(2) -> AF_INET
{-# LINE 1061 "SocketPrim.hsc" #-}

{-# LINE 1062 "SocketPrim.hsc" #-}

{-# LINE 1063 "SocketPrim.hsc" #-}
        (10) -> AF_INET6
{-# LINE 1064 "SocketPrim.hsc" #-}

{-# LINE 1065 "SocketPrim.hsc" #-}

{-# LINE 1068 "SocketPrim.hsc" #-}

{-# LINE 1071 "SocketPrim.hsc" #-}

{-# LINE 1074 "SocketPrim.hsc" #-}

{-# LINE 1077 "SocketPrim.hsc" #-}

{-# LINE 1080 "SocketPrim.hsc" #-}

{-# LINE 1083 "SocketPrim.hsc" #-}

{-# LINE 1086 "SocketPrim.hsc" #-}

{-# LINE 1089 "SocketPrim.hsc" #-}

{-# LINE 1090 "SocketPrim.hsc" #-}
	(22) -> AF_SNA
{-# LINE 1091 "SocketPrim.hsc" #-}

{-# LINE 1092 "SocketPrim.hsc" #-}

{-# LINE 1093 "SocketPrim.hsc" #-}
	(12) -> AF_DECnet
{-# LINE 1094 "SocketPrim.hsc" #-}

{-# LINE 1095 "SocketPrim.hsc" #-}

{-# LINE 1098 "SocketPrim.hsc" #-}

{-# LINE 1101 "SocketPrim.hsc" #-}

{-# LINE 1104 "SocketPrim.hsc" #-}

{-# LINE 1105 "SocketPrim.hsc" #-}
	(5) -> AF_APPLETALK
{-# LINE 1106 "SocketPrim.hsc" #-}

{-# LINE 1107 "SocketPrim.hsc" #-}

{-# LINE 1108 "SocketPrim.hsc" #-}
	(16) -> AF_ROUTE
{-# LINE 1109 "SocketPrim.hsc" #-}

{-# LINE 1110 "SocketPrim.hsc" #-}

{-# LINE 1113 "SocketPrim.hsc" #-}

{-# LINE 1116 "SocketPrim.hsc" #-}

{-# LINE 1119 "SocketPrim.hsc" #-}

{-# LINE 1122 "SocketPrim.hsc" #-}

{-# LINE 1125 "SocketPrim.hsc" #-}

{-# LINE 1128 "SocketPrim.hsc" #-}

{-# LINE 1129 "SocketPrim.hsc" #-}
	(9) -> AF_X25
{-# LINE 1130 "SocketPrim.hsc" #-}

{-# LINE 1131 "SocketPrim.hsc" #-}

{-# LINE 1132 "SocketPrim.hsc" #-}
	(3) -> AF_AX25
{-# LINE 1133 "SocketPrim.hsc" #-}

{-# LINE 1134 "SocketPrim.hsc" #-}

{-# LINE 1137 "SocketPrim.hsc" #-}

{-# LINE 1140 "SocketPrim.hsc" #-}

{-# LINE 1141 "SocketPrim.hsc" #-}
	(4) -> AF_IPX
{-# LINE 1142 "SocketPrim.hsc" #-}

{-# LINE 1143 "SocketPrim.hsc" #-}

{-# LINE 1146 "SocketPrim.hsc" #-}

{-# LINE 1149 "SocketPrim.hsc" #-}

{-# LINE 1152 "SocketPrim.hsc" #-}

{-# LINE 1155 "SocketPrim.hsc" #-}

{-# LINE 1158 "SocketPrim.hsc" #-}

{-# LINE 1161 "SocketPrim.hsc" #-}

{-# LINE 1164 "SocketPrim.hsc" #-}

{-# LINE 1167 "SocketPrim.hsc" #-}

{-# LINE 1170 "SocketPrim.hsc" #-}

{-# LINE 1173 "SocketPrim.hsc" #-}

{-# LINE 1176 "SocketPrim.hsc" #-}

{-# LINE 1179 "SocketPrim.hsc" #-}

{-# LINE 1182 "SocketPrim.hsc" #-}

{-# LINE 1185 "SocketPrim.hsc" #-}

{-# LINE 1188 "SocketPrim.hsc" #-}

{-# LINE 1191 "SocketPrim.hsc" #-}

{-# LINE 1194 "SocketPrim.hsc" #-}

{-# LINE 1197 "SocketPrim.hsc" #-}

{-# LINE 1200 "SocketPrim.hsc" #-}

{-# LINE 1203 "SocketPrim.hsc" #-}

{-# LINE 1206 "SocketPrim.hsc" #-}

-- Socket Types.

data SocketType
	= NoSocketType

{-# LINE 1212 "SocketPrim.hsc" #-}
	| Stream 

{-# LINE 1214 "SocketPrim.hsc" #-}

{-# LINE 1215 "SocketPrim.hsc" #-}
	| Datagram

{-# LINE 1217 "SocketPrim.hsc" #-}

{-# LINE 1218 "SocketPrim.hsc" #-}
	| Raw 

{-# LINE 1220 "SocketPrim.hsc" #-}

{-# LINE 1221 "SocketPrim.hsc" #-}
	| RDM 

{-# LINE 1223 "SocketPrim.hsc" #-}

{-# LINE 1224 "SocketPrim.hsc" #-}
	| SeqPacket

{-# LINE 1226 "SocketPrim.hsc" #-}
	deriving (Eq, Ord, Show)
	
packSocketType stype = case stype of
	NoSocketType -> 0

{-# LINE 1231 "SocketPrim.hsc" #-}
	Stream -> 1
{-# LINE 1232 "SocketPrim.hsc" #-}

{-# LINE 1233 "SocketPrim.hsc" #-}

{-# LINE 1234 "SocketPrim.hsc" #-}
	Datagram -> 2
{-# LINE 1235 "SocketPrim.hsc" #-}

{-# LINE 1236 "SocketPrim.hsc" #-}

{-# LINE 1237 "SocketPrim.hsc" #-}
	Raw -> 3
{-# LINE 1238 "SocketPrim.hsc" #-}

{-# LINE 1239 "SocketPrim.hsc" #-}

{-# LINE 1240 "SocketPrim.hsc" #-}
	RDM -> 4
{-# LINE 1241 "SocketPrim.hsc" #-}

{-# LINE 1242 "SocketPrim.hsc" #-}

{-# LINE 1243 "SocketPrim.hsc" #-}
	SeqPacket -> 5
{-# LINE 1244 "SocketPrim.hsc" #-}

{-# LINE 1245 "SocketPrim.hsc" #-}

-- ---------------------------------------------------------------------------
-- Utility Functions

aNY_PORT :: PortNumber 
aNY_PORT = 0

iNADDR_ANY :: HostAddress
iNADDR_ANY = htonl (0)
{-# LINE 1254 "SocketPrim.hsc" #-}

sOMAXCONN :: Int
sOMAXCONN = 128
{-# LINE 1257 "SocketPrim.hsc" #-}

maxListenQueue :: Int
maxListenQueue = sOMAXCONN

-- -----------------------------------------------------------------------------

data ShutdownCmd 
 = ShutdownReceive
 | ShutdownSend
 | ShutdownBoth

sdownCmdToInt :: ShutdownCmd -> CInt
sdownCmdToInt ShutdownReceive = 0
sdownCmdToInt ShutdownSend    = 1
sdownCmdToInt ShutdownBoth    = 2

shutdown :: Socket -> ShutdownCmd -> IO ()
shutdown (MkSocket s _ _ _ _) stype = do
  throwErrnoIfMinus1Retry "shutdown" (c_shutdown s (sdownCmdToInt stype))
  return ()

-- -----------------------------------------------------------------------------

sClose	 :: Socket -> IO ()
sClose (MkSocket s _ _ _ _) = do c_close s; return ()

-- -----------------------------------------------------------------------------

sIsConnected :: Socket -> IO Bool
sIsConnected (MkSocket _ _ _ _ status) = do
    value <- readMVar status
    return (value == Connected)	

-- -----------------------------------------------------------------------------
-- Socket Predicates

sIsBound :: Socket -> IO Bool
sIsBound (MkSocket _ _ _ _ status) = do
    value <- readMVar status
    return (value == Bound)	

sIsListening :: Socket -> IO Bool
sIsListening (MkSocket _ _ _  _ status) = do
    value <- readMVar status
    return (value == Listening)	

sIsReadable  :: Socket -> IO Bool
sIsReadable (MkSocket _ _ _ _ status) = do
    value <- readMVar status
    return (value == Listening || value == Connected)

sIsWritable  :: Socket -> IO Bool
sIsWritable = sIsReadable -- sort of.

sIsAcceptable :: Socket -> IO Bool

{-# LINE 1313 "SocketPrim.hsc" #-}
sIsAcceptable (MkSocket _ AF_UNIX Stream _ status) = do
    value <- readMVar status
    return (value == Connected || value == Bound || value == Listening)
sIsAcceptable (MkSocket _ AF_UNIX _ _ _) = return False

{-# LINE 1318 "SocketPrim.hsc" #-}
sIsAcceptable (MkSocket _ _ _ _ status) = do
    value <- readMVar status
    return (value == Connected || value == Listening)
    
-- -----------------------------------------------------------------------------
-- Internet address manipulation routines:

inet_addr :: String -> IO HostAddress
inet_addr ipstr = do
   withCString ipstr $ \str -> do
   had <- c_inet_addr str
   if had == -1
    then ioError (userError ("inet_addr: Malformed address: " ++ ipstr))
    else return had  -- network byte order

inet_ntoa :: HostAddress -> IO String
inet_ntoa haddr = do
  pstr <- c_inet_ntoa haddr
  peekCString pstr

-- socketHandle turns a Socket into a Haskell IO Handle. By default, the new
-- handle is unbuffered. Use hSetBuffering to alter this.


{-# LINE 1342 "SocketPrim.hsc" #-}
socketToHandle :: Socket -> IOMode -> IO Handle
socketToHandle s@(MkSocket fd _ _ _ _) mode = do
    openFd (fromIntegral fd) (Just PrelPosix.Stream) (show s) mode True{-bin-} False{-no truncate-}

{-# LINE 1349 "SocketPrim.hsc" #-}

-- ---------------------------------------------------------------------------
-- WinSock support

-- If you're using WinSock, the programmer has to call a startup
-- routine before starting to use the goods. So, if you want to
-- stay portable across all ghc-supported platforms, you have to
-- use @withSocketsDo@...:

withSocketsDo :: IO a -> IO a

{-# LINE 1360 "SocketPrim.hsc" #-}
withSocketsDo x = x

{-# LINE 1375 "SocketPrim.hsc" #-}

-- ---------------------------------------------------------------------------
-- foreign imports from the C library

foreign import "my_inet_ntoa" unsafe
  c_inet_ntoa :: HostAddress -> IO (Ptr CChar)

{-# LINE 1383 "SocketPrim.hsc" #-}

foreign import "inet_addr" unsafe
  c_inet_addr :: Ptr CChar -> IO HostAddress

foreign import "shutdown" unsafe
  c_shutdown :: CInt -> CInt -> IO CInt 
foreign import "close" unsafe
  c_close :: CInt -> IO CInt

foreign import "socket" unsafe
  c_socket :: CInt -> CInt -> CInt -> IO CInt
foreign import "bind" unsafe
  c_bind :: CInt -> Ptr SockAddr -> CInt{-CSockLen???-} -> IO CInt
foreign import "connect" unsafe
  c_connect :: CInt -> Ptr SockAddr -> CInt{-CSockLen???-} -> IO CInt
foreign import "accept" unsafe
  c_accept :: CInt -> Ptr SockAddr -> Ptr CInt{-CSockLen???-} -> IO CInt
foreign import "listen" unsafe
  c_listen :: CInt -> CInt -> IO CInt

foreign import "write" unsafe
  c_write :: CInt -> Ptr CChar -> CSize -> IO CInt
foreign import "read" unsafe
  c_read :: CInt -> Ptr CChar -> CSize -> IO CSize
foreign import "send" unsafe
  c_send :: CInt -> Ptr CChar -> CSize -> CInt -> IO CInt
foreign import "sendto" unsafe
  c_sendto :: CInt -> Ptr CChar -> CSize -> CInt -> Ptr SockAddr -> CInt -> IO CInt
foreign import "recv" unsafe
  c_recv :: CInt -> Ptr CChar -> CSize -> CInt -> IO CInt
foreign import "recvfrom" unsafe
  c_recvfrom :: CInt -> Ptr CChar -> CSize -> CInt -> Ptr SockAddr -> Ptr CInt -> IO CInt
foreign import "getpeername" unsafe
  c_getpeername :: CInt -> Ptr SockAddr -> Ptr CInt -> IO CInt
foreign import "getsockname" unsafe
  c_getsockname :: CInt -> Ptr SockAddr -> Ptr CInt -> IO CInt

foreign import "getsockopt" unsafe
  c_getsockopt :: CInt -> CInt -> CInt -> Ptr CInt -> Ptr CInt -> IO CInt
foreign import "setsockopt" unsafe
  c_setsockopt :: CInt -> CInt -> CInt -> Ptr CInt -> CInt -> IO CInt

-----------------------------------------------------------------------------
-- Support for thread-safe blocking operations in GHC.


{-# LINE 1429 "SocketPrim.hsc" #-}

{-# SPECIALISE 
    throwErrnoIfMinus1Retry_mayBlock
	 :: String -> IO CInt -> IO CInt -> IO CInt #-}
throwErrnoIfMinus1Retry_mayBlock :: Num a => String -> IO a -> IO a -> IO a
throwErrnoIfMinus1Retry_mayBlock name on_block act = do
    res <- act
    if res == -1
        then do
            err <- getErrno
            if err == eINTR
                then throwErrnoIfMinus1Retry_mayBlock name on_block act
	        else if err == eWOULDBLOCK || err == eAGAIN
		        then on_block
                        else throwErrno name
        else return res

throwErrnoIfMinus1Retry_repeatOnBlock :: Num a => String -> IO b -> IO a -> IO a
throwErrnoIfMinus1Retry_repeatOnBlock name on_block act = do
  throwErrnoIfMinus1Retry_mayBlock name (on_block >> repeat) act
  where repeat = throwErrnoIfMinus1Retry_repeatOnBlock name on_block act

setSocketNonBlocking fd = do
  flags <- fcntl_read fd (3)
{-# LINE 1453 "SocketPrim.hsc" #-}
  fcntl_write fd (4) (flags .|. 2048)
{-# LINE 1454 "SocketPrim.hsc" #-}

foreign import "fcntl" unsafe fcntl_read  :: CInt -> CInt -> IO CInt
foreign import "fcntl" unsafe fcntl_write :: CInt -> CInt -> CInt -> IO CInt

{-# LINE 1469 "SocketPrim.hsc" #-}

